/**
 *
 * This file is part of Tulip (www.tulip-software.org)
 *
 * Authors: David Auber, Patrick Mary, Morgan Mathiaut
 * from the LaBRI Visualization Team
 * Last modification : 15/01/2010
 *
 * Tulip is free software; you can redistribute it and/or modify
 * it under the terms of the Lesser GNU General Public License
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * Tulip is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include "tulip/Types.h"
#include "tulip/AbstractProperty.h"
#include "tulip/Color.h"
#include "tulip/DoubleProperty.h"
#include "tulip/StringProperty.h"
#include "tulip/BooleanProperty.h"
#include "tulip/LayoutProperty.h"
#include "tulip/IntegerProperty.h"
#include "tulip/ColorProperty.h"
#include "tulip/SizeProperty.h"

#include "tulip/Reflect.h"

using namespace std;
using namespace tlp;

DataSet::DataSet(const DataSet &set) {
  *this = set;
}

DataSet & DataSet::operator=(const DataSet &set) {
  if (this != &set) {
    data.clear();
    for (std::list< std::pair<std::string, tlp::DataType*> >::const_iterator it =
	   set.data.begin(); it != set.data.end(); ++it) {
      data.push_back(std::pair<std::string, tlp::DataType*>((*it).first, (*it).second->clone()));
    }
  }
  return *this;
}

DataSet::~DataSet() {
  for (std::list< std::pair<std::string, tlp::DataType*> >::iterator it =
	 data.begin(); it != data.end(); ++it) {
    if (it->second)
      delete it->second;
  }
}

bool DataSet::exist(const string &str) const {
  for (std::list< std::pair<std::string, tlp::DataType*> >::const_iterator it =
	 data.begin(); it != data.end(); ++it) {
    if ((*it).first == str)
      return true;
  }
  return false;
}

void DataSet::remove(const string &str) {
  for (std::list< std::pair<std::string, tlp::DataType*> >::iterator it =
	 data.begin(); it != data.end(); ++it) {
    if ((*it).first == str) {
      if (it->second)
	delete it->second;
      data.erase(it);
      break;
    }
  }
}

DataType* DataSet::getData(const string &str) const {
  for (std::list< std::pair<std::string, tlp::DataType*> >::const_iterator it =
	 data.begin(); it != data.end(); ++it) {
    if ((*it).first == str)
      return it->second ? it->second->clone() : NULL;
  }
  return NULL;
}

void DataSet::setData(const std::string &str, const DataType* value) {
  DataType* val = value ? value->clone() : NULL;
  for (std::list< std::pair<std::string, tlp::DataType*> >::iterator it =
	 data.begin(); it != data.end(); ++it) {
    std::pair<std::string, tlp::DataType*> &p = *it;
    if (p.first == str) {
      if (p.second)
	delete p.second;
      p.second = val;
      return;
    }
  }
  data.push_back(std::pair<std::string, tlp::DataType*>(str, val));
}

Iterator< pair<string, DataType*> >* DataSet::getValues() const {
  list< pair<string, DataType*> >::const_iterator begin = data.begin();
  list< pair<string, DataType*> >::const_iterator end = data.end();
  
  return new StlIterator<pair<string, DataType*>, list< pair<string, DataType*> >::const_iterator>(begin, end);
}

bool StructDef::hasField(string str) {
  for (std::list< std::pair<std::string, std::string> >::iterator it =
	 data.begin(); it != data.end(); ++it) {
    if ((*it).first == str)
      return true;
  }
  return false;
}

void StructDef::erase(string str) {
  for (std::list< std::pair<std::string, std::string> >::iterator it =
	 data.begin(); it != data.end(); ++it) {
    if ((*it).first == str) {
      data.erase(it);
      break;
    }
  }
  help.erase( help.find(str) );
  defValue.erase( defValue.find(str) );
}

Iterator< pair<string, string> >* StructDef::getField() const {
  list< pair<string, string> >::const_iterator begin = data.begin();
  list< pair<string, string> >::const_iterator end = data.end();
  
  return new StlIterator<pair<string, string>, list< pair<string, string> >::const_iterator>(begin, end);
}

string StructDef::getHelp( string str ) const {
  map<string,string>::const_iterator it = help.find(str);
  if( it == help.end() )
  	return string();
  else
  	return it->second;
}

string StructDef::getDefValue( string str ) const {
  map<string,string>::const_iterator it = defValue.find(str);
  if( it == defValue.end() )
    return string();
  else
    return it->second;
}

void StructDef::setDefValue(string name, string value) {
  map<string,string>::iterator it = defValue.find(name);
  if( it == defValue.end() )
    defValue[name] = value;
  else
    it->second = value;
}

bool StructDef::isMandatory( string str ) const {
  map<string,bool>::const_iterator it = mandatory.find(str);
  if( it == mandatory.end() )
    return false;
  else
    return it->second;
}

#undef TN
#define	TN( T )	typeid(T).name()

void StructDef::buildDefaultDataSet(DataSet &ioDataSet, Graph *inG) {
  Iterator< pair<string,string> > * defIt;
  defIt = getField();

  while( defIt->hasNext() ) {
    pair<string,string> def;
    def = defIt->next();
    const string & name  = def.first;
    const string & tname = def.second;
    // Already defined ?
    if( ioDataSet.exist(name) )
      continue;
    // Has def value ?
    const string & defv = getDefValue( name );
    if( defv.size() == 0 )
      continue;
    // bool
    if( tname == TN(bool) ){
      bool v;
      bool res = BooleanType::fromString( v, defv );
      assert( res );
      if( res )
	ioDataSet.set( name, v );
    }
    // int
    else if( tname == TN(int) ) {
      int v;
      bool res = IntegerType::fromString( v, defv );
      assert( res );
      if( res )
	ioDataSet.set( name, v );
    }
    // unsigned int
    else if( tname == TN(unsigned int) ) {
      unsigned int v = atol( defv.c_str() );
      ioDataSet.set( name, v );
    }
    // unsigned int
    else if( tname == TN(long) ) {
      long v = atol( defv.c_str() );
      ioDataSet.set( name, v );
    }
    // fp
    else if( tname == TN(double) ) {
      double v;
      bool res = DoubleType::fromString( v, defv );
      assert( res );
      if( res )
	ioDataSet.set( name, (double)v );
    }
    else if( tname == TN(float) ) {
      double v;
      bool res = DoubleType::fromString( v, defv );
      assert( res );
      if( res )
	ioDataSet.set( name, (float)v );
    }
    // string
    else if( tname == TN(string) ) {
      ioDataSet.set( name, defv );
    }
    // Color
    else if( tname == TN(Color) ) {
      Color v;
      bool res = ColorType::fromString( v, defv );
      assert( res );
      if( res ) {
	ioDataSet.set( name, v );
	//cout << v;
      }
    }
    // Size
    else if( tname == TN(Size) ) {
      Size v;
      bool res = SizeType::fromString( v, defv );
      assert( res );
      if( res )
	ioDataSet.set( name, v );
    } else
      // look for an already existing property
      if (inG && (inG->existProperty(defv))) {
	// BooleanProperty
	if( tname == TN(BooleanProperty) ) {
	  ioDataSet.set( name, inG->getProperty<BooleanProperty>(defv) );
	}
	// DoubleProperty
	else if( tname == TN(DoubleProperty) ) {
	  // look for an already existing property
	  if (inG->existProperty(defv))
	    ioDataSet.set( name, inG->getProperty<DoubleProperty>(defv) );
	}
	// LayoutProperty
	else if( tname == TN(LayoutProperty) ) {
	  ioDataSet.set( name, inG->getProperty<LayoutProperty>(defv) );
	}
	// StringProperty
	else if( tname == TN(StringProperty) ) {
	  ioDataSet.set( name, inG->getProperty<StringProperty>(defv) );
	}
	// IntegerProperty
	else if( tname == TN(IntegerProperty) ) {
	  ioDataSet.set( name, inG->getProperty<IntegerProperty>(defv) );
	}
	// SizeProperty
	else if( tname == TN(SizeProperty) ) {
	  ioDataSet.set( name, inG->getProperty<SizeProperty>(defv) );
	}
	// ColorProperty
	else if( tname == TN(ColorProperty) ) {
	  ioDataSet.set( name, inG->getProperty<ColorProperty>(defv) );
	}
	//
	else if ( tname == TN(PropertyInterface*) ) {
	  ioDataSet.set( name, inG->getProperty(defv) );
	}
      }
  }
  delete defIt;
}
